/*
  File:	wffConvHV.c
  Authors: K.R. Sloan
  Last Modified: 3 October 1990
  Purpose: convolves a wff image with a Mx1 template,
           plus a constant (DC) value;
           followed by a 1xN template, plus a constant (DC value.

           The image is treated as if it were embedded in a 
           field of half-intensity values.  This avoids
           one type of edge effect, but introduces another. 

           Template (and DC) values are scaled integers
           (positive or negative.)  The scale factor is 2048.  So,
            2048 == 1.0, 1024 == 0.5, -512 == -0.25, etc.
           
           It is strongly recommended that the individual magnitudes not
           be too large, and that the sum of the values in the template
           (plus the DC value) lie in the range [0,2048] (i.e., [0.0,1.0])
           
           It is alwo strongly recommended that the individual filters
           have an ODD size.  If the filter width is EVEN, then your image
           will be shifted, as well as filtered!

 */

#include <stdio.h>
#include <wff.h>
#include <sys/file.h>
#include <sys/errno.h>

#define ShiftFactor 11
#define ScaleFactor 2048

static char RoutineName[] = "wffConvHV";

static int ScanLinesRead;
static int ReadAndFilter(FBin,RawLine,Scanline,width,hn,h,hBorder,
                              BandsPerPixel,Gray,MaxPixelValue)
 FrameBufferType *FBin;
 unsigned short *RawLine;
 unsigned short *Scanline;
 int width;
 int hn;
 int h[];
 int hBorder;
 int BandsPerPixel;
 unsigned short Gray, MaxPixelValue;
 {
  int x,b,sx;
  unsigned short *in, *out, *t;
  unsigned int value;

  t = RawLine;
  for(x=0;x<hBorder;x++)
   for(b=0;b<BandsPerPixel;b++) *t++ = Gray;

  if (FAILURE == NextNPixelsIn(FBin,width,t)) return(FAILURE);

  for(x += width, t += width*BandsPerPixel;
      x < (width+hn);
      x++)
   for(b=0;b<BandsPerPixel;b++) *t++ = Gray;

  out = Scanline;
  for (x=0;x<width;x++)
   for (b=0;b<BandsPerPixel;b++)
    {
     int *K;

     in = RawLine+(x*BandsPerPixel)+b;
     K = h;
     value = (*K++)*(MaxPixelValue);  /* dc */
     for (sx=1;sx<hn;sx++,in += BandsPerPixel,K++)
      value += (int)(*K) * (int)(*in);

     value >>= ShiftFactor;
     if      (value < 0)             value = 0;
     else if (value > MaxPixelValue) value = MaxPixelValue;
     *out++ = (unsigned short) value;
    }
  return(SUCCESS);
 }


int
wffConvHV(fdIn, fdOut, hn, vn, h, v)
 FILE *fdIn, *fdOut;
 int hn, vn;
 int h[], v[];
 {
  FrameBufferType *FBin, *FBout;
  int Bottom, Left, Top, Right, BandsPerPixel, passed;
  char WhatBands[10];
  int BitsPerBand;
  char Name[NameLength], Value[ValueLength];
  unsigned short *t;
  unsigned short *RawLine;
  unsigned short **ScanLine;
  unsigned short *NewLine;
  int MaxPixelValue, width, height;
  unsigned short Gray;
  int n;
  int x, y, sx, sy, b;
  int hBorder, vBorder;

  FBin  = (FrameBufferType *)0;
  FBout = (FrameBufferType *)0;
  if (FAILURE == OpenFB(&FBin))          {                     return FAILURE;}
  if (FAILURE == PassImageIn(fdIn, FBin)){(void)CloseFB(&FBin);return FAILURE;}

  if (FAILURE == GetBounds(FBin, &Bottom, &Left, &Top, &Right)) 
   { (void)CloseFB(FBin); return FAILURE; }
  if (FAILURE == GetColorSystem(FBin, WhatBands, &BitsPerBand))
   { (void)CloseFB(FBin); return FAILURE; }
  if (FAILURE == GetDescriptor(FBin,"BitsPerBand",Value))
   { (void)CloseFB(FBin); return FAILURE; }
  BitsPerBand = atoi(Value);
  BandsPerPixel = strlen(WhatBands);

  if (FAILURE == OpenFB(&FBout))         {(void)CloseFB(&FBin);return FAILURE;}

  /*  Copy over existing NV pairs - watch for "X-PassedBy" */
  passed = 0;
  for (n=0;;n++)
   {
    GetDescriptorN(FBin, n, Name, Value);
    if (Name[0] == '\0') break;
    if (0 == strcmp(Name,"X-PassedBy"))
     {
      if ( (strlen(Value)+strlen(RoutineName)+3) > ValueLength)
       strcpy(Value,"...");
      strcat(Value,", "); strcat(Value,RoutineName);
      passed = 1;
     }
    SetDescriptor(FBout, Name, Value);
   }

  /*  if necessary, add "X-PassedBy" */
  if (0 == passed)
   {
    strcpy(Name,"X-PassedBy");
    strcpy(Value,RoutineName);
    SetDescriptor(FBout, Name, Value);
   }

  /* Header operations over, now we can start the output stream */
  if (FAILURE == PassImageOut(fdOut, FBout))
   { (void)CloseFB(&FBin); (void)CloseFB(&FBout); return 1; }

  /* Finally, pass the pixels */

  width = Right-Left+1;
  height = Top-Bottom+1;
  MaxPixelValue = (1 << BitsPerBand) - 1;
  Gray = MaxPixelValue >> 1;

  /*
    if hn,vn are even, shift the image down and to the right
    this is a completely arbitrary choice.
   */
  hBorder = (hn-1) >> 1; vBorder = (vn-1) >> 1;

  /*
    allocate scanlines and fill with gray
   */
  NewLine = (unsigned short *) malloc(width*BandsPerPixel
                                      * sizeof (unsigned short));
  if ((unsigned short *)0 == NewLine)
   {
    fprintf(stderr,"%s: cannot allocate NewLine\n",
                   RoutineName);
    (void)CloseFB(&FBin); (void)CloseFB(&FBout); return FAILURE;
   }
  

  ScanLine = (unsigned short **) calloc(vn-1,
                                        sizeof (unsigned short *));
  if ((unsigned short **)0 == ScanLine)
    {
    fprintf(stderr,"%s: cannot allocate Scanline\n",
                   RoutineName);
    free(NewLine);
    (void)CloseFB(&FBin); (void)CloseFB(&FBout); return FAILURE;
  }

  for(sy=0;sy<(vn-1);sy++)
   {
    ScanLine[sy] = (unsigned short *) malloc(width*BandsPerPixel
                                             * sizeof (unsigned short));
    if ((unsigned short *)0 == ScanLine[sy])
     {
      fprintf(stderr,"%s: cannot allocate Scanline[%d]\n",
                     RoutineName,sy);
      { int fy; for (fy = 0; fy < sy; fy++) free(ScanLine[fy]); }
      free(NewLine); free(ScanLine);
      (void)CloseFB(&FBin); (void)CloseFB(&FBout); return FAILURE;
     }
    t = ScanLine[sy];
    for(sx=0;sx<width;sx++)
     for (b=0;b<BandsPerPixel;b++) *t++ = Gray;
   }

  RawLine = (unsigned short *) calloc(width+hn-1,
                                      sizeof (unsigned short *));
  if ((unsigned short *)0 == RawLine)
   {
    fprintf(stderr,"%s: cannot allocate Rawline\n",
                   RoutineName);
    { int fy; for (fy = 0; fy < (vn-1); fy++) free(ScanLine[fy]); }
    free(NewLine); free(ScanLine);
    (void)CloseFB(&FBin); (void)CloseFB(&FBout); return FAILURE;
  }

  /* preload the pipeline */
  ScanLinesRead = 0;
  for(sy=vBorder+1,            ScanLinesRead = 0;
      sy<(vn-1);
      sy++,                    ScanLinesRead++)
   if (FAILURE == ReadAndFilter(FBin,RawLine,ScanLine[sy],width,hn,h,hBorder,
                                BandsPerPixel,Gray,MaxPixelValue))
    {
     fprintf(stderr,"%s: failed reading Scanline %d\n",
                    RoutineName, ScanLinesRead);
     { int fy; for (fy = 0; fy < (vn-1); fy++) free(ScanLine[fy]); }
     free(NewLine); free(ScanLine); free(RawLine);
     (void)CloseFB(&FBin); (void)CloseFB(&FBout); return FAILURE; 
    }

  /* step through destination, one scanline at a time */
  for (y = Bottom; y <= Top; y++)
   {
    /* move forward one line */
    t = ScanLine[0];
    for(sy=1;sy<(vn-1);sy++) ScanLine[sy-1] = ScanLine[sy];
    ScanLine[vn-2] = t;

    if (ScanLinesRead < height)
     {
      if (FAILURE == ReadAndFilter(FBin,RawLine,t,width,hn,h,hBorder,
                                   BandsPerPixel,Gray,MaxPixelValue)) 
       {
        fprintf(stderr,"%s: failed reading Scanline %d\n",
                        RoutineName, ScanLinesRead);
        { int fy; for (fy = 0; fy < (vn-1); fy++) free(ScanLine[fy]); }
        free(NewLine); free(ScanLine); free(RawLine);
        (void)CloseFB(&FBin); (void)CloseFB(&FBout); return FAILURE;
       }
     ScanLinesRead++;
    }
    else
     {  /* last lines - pad with gray */
      for(sx=0;sx<width;sx++)
       for (b=0;b<BandsPerPixel;b++) *t++ = Gray;
     }
    /* filter vertically */
    {
     register int value;
     register int *k;
     unsigned short *out;

     sx = 0;
     out = NewLine;
     for (x = Left; x <= Right; x++)
      for (b=0;b<BandsPerPixel; b++,sx++,out++)
       {
        k = v;
        value = (*(k++))*MaxPixelValue;  /* dc */
        for (sy=0;sy<(vn-1);sy++)
         value += (*(k++))*ScanLine[sy][sx];
        value >>= ShiftFactor;
        if      (value < 0)             value = 0;
        else if (value > MaxPixelValue) value = MaxPixelValue;
        *out = (unsigned short) value;
       }
    }
    if (FAILURE == NextNPixelsOut(FBout,width,NewLine)) 
     {
      { int fy; for (fy = 0; fy < (vn-1); fy++) free(ScanLine[fy]); }
      free(NewLine); free(ScanLine); free(RawLine);
      (void)CloseFB(&FBin); (void)CloseFB(&FBout); return FAILURE;
     }
   }

  { int fy; for (fy = 0; fy < (vn-1); fy++) free(ScanLine[fy]); }
  free(NewLine); free(ScanLine); free(RawLine);
  (void)CloseFB(&FBin);
  return (CloseFB(&FBout));

 }
